import pathway as pw
from llm_app.model_wrappers.api_clients.clients import (
    OpenAIChatCompletionClient,
    OpenAIClient,
    OpenAIEmbeddingClient,
)
from llm_app.model_wrappers.base import APIModel


class MessagePreparer:
    @staticmethod
    def prepare_chat_messages(prompt: str):
        return [
            dict(role="system", content="You are a helpful assistant"),
            dict(role="user", content=prompt),
        ]


class OpenAIChatGPTModel(APIModel):
    def get_client(self, openai_key: str) -> OpenAIClient:
        return OpenAIChatCompletionClient(openai_key)

    def __call__(self, text: str, locator="gpt-3.5-turbo", **kwargs) -> str:
        """

        Example

        # >>> model = OpenAIChatGPTModel(api_key = api_key)
        # >>> model(
        # ...     locator='gpt-4-0613',
        # ...     text="Tell me a joke about jokes",
        # ...     temperature=1.1
        # ... )
        """
        messages = MessagePreparer.prepare_chat_messages(text)
        response = self.api_client.make_request(
            messages=messages, model=locator, **kwargs
        )
        return response.choices[0].message.content

    def apply(
        self,
        *args,
        **kwargs,
    ) -> pw.ColumnExpression:
        """
        Applies the specified model in `locator` from OpenAIChatGPT API to the provided text.
        Parameters
        ----------
        text : Union[pw.ColumnExpression, str]
            The input text on which the model will be applied. It can be a column expression or a string.
        locator : Union[pw.ColumnExpression, str, None]
            The model locator to use for applying the model.
            If provided, it should be a column expression or a string.
            Otherwise, the default chat completion model `gpt-3.5-turbo` is applied.
            Please check out https://platform.openai.com/docs/models/model-endpoint-compatibility
            to see the available models.
        **kwargs : dict
            Additional keyword arguments that will be used for the model application.
            These could include settings such as `temperature`, `max_tokens`, etc.
            Check https://platform.openai.com/docs/api-reference/chat/create for the official API Reference
        Returns
        -------
        pw.ColumnExpression
            The result of the model application as a column expression or str.
            Please note that the output is `chat_completion.choices[0].message.content`
            where `chat_completion` is the api response.
        Example:
        # >>> model = OpenAIChatGPTModel(api_key = api_key)
        # >>>
        # >>> table = pw.debug.table_from_pandas(
        # ...     pd.DataFrame.from_records([
        # ...         {"text": "How to use pathway to process a kafka stream ?"},
        # ...         {"text": "How to apply a function to a pathway table ?"}
        # ...     ])
        # ... )
        # >>> table += table.select(
        # ...     response = model.apply(
        # ...         pw.this.text,
        # ...         locator='gpt-4',
        # ...         temperature=1.5,
        # ...         max_tokens=1000
        # ...     )
        # ... )
        """
        return super().apply(*args, **kwargs)


class OpenAIEmbeddingModel(APIModel):
    def get_client(self, openai_key: str) -> OpenAIClient:
        return OpenAIEmbeddingClient(openai_key)

    def __call__(self, text: str, locator="text-embedding-ada-002", **kwargs):
        """
        Example:

        # >>> embedder = OpenAIEmbeddingModel(api_key)
        # >>>
        # >>> embedder(
        # ...    text='Some random text'
        # ...    locator='text-embedding-ada-002'
        # ... )
        """

        response = self.api_client.make_request(input=[text], model=locator, **kwargs)
        return response["data"][0]["embedding"]

    def apply(
        self,
        *args,
        **kwargs,
    ) -> pw.ColumnExpression:
        """
        Applies the specified model in `locator` from OpenAIEmbeddingModel API to the provided text.
        Parameters
        ----------
        text : Union[pw.ColumnExpression, str]
            The input text on which the model will be applied. It can be a column expression or a constant value.
        locator : Union[pw.ColumnExpression, str, None]
            The model locator to use for applying the model.
            If provided, it should be a column expression or a constant value.
            Otherwise, the default chat completion model `gpt-3.5-turbo` is applied.
            Please check out https://platform.openai.com/docs/models/model-endpoint-compatibility
            to see the available models.
        **kwargs : dict
            Additional keyword arguments that will be used for the model application.
            These could include settings such as `temperature`, `max_tokens`, etc.
            You can check https://platform.openai.com/docs/api-reference/embeddings/create
            for the official API Reference.
        Returns
        -------
        pw.ColumnExpression
            The result of the model application as a column expression or constant of type list.
            Please note that the output is `results["data"][0]["embedding"]`
        Example:
        # >>> embedder = OpenAIEmbeddingModel(api_key)
        # >>>
        # >>> table = pw.debug.table_from_pandas(
        # ...     pd.DataFrame.from_records([
        # ...         {"text": "How to use pathway to process a kafka stream ?"},
        # ...         {"text": "How to apply a function to a pathway table ?"}
        # ...     ])
        # ... )
        # >>> table += table.select(
        # ...     embedding = embedder.apply(
        # ...         pw.this.text,
        # ...         locator='text-embedding-ada-002'
        # ...     )
        # ... )
        """
        return super().apply(*args, **kwargs)



import logging
from abc import ABC, abstractmethod
import requests


logfun = logging.debug


class APIClient(ABC):
    @abstractmethod
    def make_request(self, **kwargs):
        pass


class OpenAIClient(APIClient):
    def __init__(self, api_key: str):
        import openai
        self.api = openai
        self.api.api_key = api_key


class OpenAIChatCompletionClient(OpenAIClient):
    def make_request(self, **kwargs):
        logfun("Calling OpenAI chat completion service %s", str(kwargs)[:100])
        return self.api.ChatCompletion.create(**kwargs)


class OpenAIEmbeddingClient(OpenAIClient):
    def make_request(self, **kwargs):
        logfun("Calling OpenAI embedding service %s", str(kwargs)[:100])
        return self.api.Embedding.create(**kwargs)


class HuggingFaceClient(APIClient):
    def __init__(self, api_key: str) -> None:
        self.headers = {"Authorization": f"Bearer {api_key}"}
        self.api_url_prefix = "https://api-inference.huggingface.co/models"

    def make_request(self, **kwargs):
        logfun("Calling HuggingFace %s", str(kwargs)[:100])
        endpoint = kwargs.pop("model")
        url = f"{self.api_url_prefix}/{endpoint}"
        response = requests.post(url, headers=self.headers, json=kwargs)
        return response.json()



from llm_app.model_wrappers.base import BaseModel


class HFPipelineTask(BaseModel):
    def __init__(self, model, device="cpu", **kwargs):
        """
        A wrapper class for Hugging Face's `Pipeline` class.

        The `pipeline` function from Hugging Face is a utility factory method that creates
        a Pipeline to handle different tasks.
        It supports tasks like text classification, translation, summarization, and many more.

        This wrapper class simplifies the process of initializing the pipeline and allows the user
        to easily change the underlying model used for computations.

        Parameters:
        -----------
        model : str, required
            The model identifier from Hugging Face's model hub.
        device : str, default='cpu'
            The device where the computations will be performed.
            Supports 'cpu' or 'gpu'. Default is 'cpu'.
        **kwargs : optional
            Additional arguments form HF.
            Please check out https://huggingface.co/docs/transformers/main/main_classes/pipelines
            for more information on the models and available arguments.

        Attributes:
        -----------
        pipeline : transformers.Pipeline
            The Hugging Face pipeline object.
        tokenizer : transformers.PreTrainedTokenizer
            The tokenizer associated with the pipeline.

        Example:
        --------
        >>> pipe = HFPipelineTask('gpt2')
        >>> result = pipe('Hello world')
        """
        from transformers import pipeline

        super().__init__(**kwargs)
        self.pipeline = pipeline(model=model, device=device)
        self.tokenizer = self.pipeline.tokenizer

    def crop_to_max_length(self, input_string, max_length=500):
        tokens = self.tokenizer.tokenize(input_string)
        if len(tokens) > max_length:
            tokens = tokens[:max_length]
        return self.tokenizer.convert_tokens_to_string(tokens)


class HFFeatureExtractionTask(HFPipelineTask):
    def __init__(self, model, device="cpu", max_length=500, **kwargs):
        super().__init__(model, device=device, **kwargs)
        self.max_length = max_length

    def __call__(self, text, **kwargs):
        """
        This method computes feature embeddings for the given text.
        HuggingFace Feature extraction models return embeddings per token.
        To get the embedding vector of a text, we simply take the average.

        Args:
            text (str): The text for which we compute the embedding.
            **kwargs: Additional arguments to be passed to the pipeline.

        Returns:
            List[float]: The average feature embeddings computed by the model.
        """

        text = self.crop_to_max_length(text, max_length=self.max_length)
        # This will return a list of lists (one list for each word in the text)
        embedding = self.pipeline(text, **kwargs)[0]

        # For simplicity, we'll just average all word vectors to get a sentence embedding
        avg_embedding = [sum(col) / len(col) for col in zip(*embedding)]

        return avg_embedding


class HFTextGenerationTask(HFPipelineTask):
    def __init__(
        self, model, device="cpu", max_prompt_length=500, max_new_tokens=500, **kwargs
    ):
        super().__init__(model, device=device, **kwargs)
        self.max_prompt_length = max_prompt_length
        self.max_new_tokens = max_new_tokens

    def __call__(self, text, **kwargs):
        """
        Run the model to complete the text.
        Args:
            text (str): prompt to complete.
            return_full_text (bool, optional, defaults to True):
                If True, returns the full text, if False, only added text is returned.
                Only significant if return_text is True.
            clean_up_tokenization_spaces (bool, optional, defaults to False):
                If True, removes extra spaces in text output.
            prefix (str, optional): Adds prefix to prompt.
            handle_long_generation (str, optional): By default, doesn't handle long generation.
                Provides strategies to address this based on your use case:
                    None: Does nothing special
                    "hole": Truncates left of input, leaving a gap for generation.
                        Might truncate a lot of the prompt, not suitable when generation exceeds model capacity.
            Other arguments from transformers.TextGenerationPipeline.__call__ are supported as well. Link:
            https://huggingface.co/docs/transformers/main/main_classes/pipelines#transformers.TextGenerationPipeline.__call__

        """
        text = self.crop_to_max_length(text, self.max_prompt_length)

        max_new_tokens = kwargs.pop("max_new_tokens", self.max_new_tokens)

        output = self.pipeline(text, max_new_tokens=max_new_tokens, **kwargs)
        return output[0]["generated_text"]



from llm_app.model_wrappers.huggingface_wrapper.api_models import (
    HFApiFeatureExtractionTask,
    HFApiTextGenerationTask,
)
from llm_app.model_wrappers.huggingface_wrapper.pipelines import (
    HFFeatureExtractionTask,
    HFTextGenerationTask,
)
from llm_app.model_wrappers.openai_wrapper.api_models import (
    OpenAIChatGPTModel,
    OpenAIEmbeddingModel,
)
from llm_app.model_wrappers.sentence_transformer.embedding import (
    SentenceTransformerTask,
)

__all__ = [
    "HFApiFeatureExtractionTask",
    "HFApiTextGenerationTask",
    "HFFeatureExtractionTask",
    "HFTextGenerationTask",
    "OpenAIChatGPTModel",
    "OpenAIEmbeddingModel",
    "SentenceTransformerTask",
]



from llm_app.model_wrappers.base import BaseModel


class SentenceTransformerTask(BaseModel):
    def __init__(self, model: str, **kwargs):
        """
        Wrapper for sentence-transformers.
        Arguments:
            model: model name or path
        """
        from sentence_transformers import SentenceTransformer

        super().__init__(**kwargs)
        self.model = SentenceTransformer(model_name_or_path=model)

    def __call__(self, text: str, **kwargs) -> str:
        """
        Arguments:
            text: input text string.
            **kwargs: Check out https://www.sbert.net/
        """
        return self.model.encode(text, **kwargs).tolist()



from llm_app.model_wrappers.api_clients.clients import HuggingFaceClient
from llm_app.model_wrappers.base import APIModel


class HuggingFaceAPIModel(APIModel):
    def get_client(self, api_key: str) -> HuggingFaceClient:
        return HuggingFaceClient(api_key=api_key)

    def call_api(self, **kwargs):
        """
        Makes a request to the Hugging Face API and returns the result.

        The method accepts arguments as keyword arguments (**kwargs).
        The expected arguments are 'model' and others that depend on the specific task.
        Please check [HuggingFace Inference API](https://huggingface.co/docs/api-inference/detailed_parameters)
        'model' is a string representing the pre-trained model to use for the call.

        Examples:

        1) Question-Answering Task usage:

        model = HuggingFaceModel(api_key=token)
        result = model.call(
            inputs=dict(
                context="Pathway is a realtime stream data processing framework. It has a python api",
                question="Does Pathway have a python API ?"
            ),
            model='deepset/roberta-base-squad2'
        )
        The expected output for the example above is:
        {
            'score': 0.41046786308288574,
            'start': 56,
            'end': 75,
            'answer': 'It has a python api'
        }

        2) Sentiment analysis model usage:

        result = model.call(
            inputs="What a performance that was!",
            model='distilbert-base-uncased-finetuned-sst-2-english'
        )
        The expected output for the example above is:
        [
            [
                {'label': 'POSITIVE', 'score': 0.9988183379173279},
                {'label': 'NEGATIVE', 'score': 0.0011816363548859954}
            ]
        ]

        Args:
            **kwargs: The arguments for the model call. 'inputs' and 'model' keys are expected.

        Returns:
            The response from the Hugging Face API, the format depends on the model being used.
        """
        return self.api_client.make_request(**kwargs)


class HFApiFeatureExtractionTask(HuggingFaceAPIModel):
    def __call__(
        self, text: str, locator="sentence-transformers/all-MiniLM-L6-v2", **kwargs
    ):
        response = self.call_api(inputs=[text], model=locator, **kwargs)
        return response


class HFApiTextGenerationTask(HuggingFaceAPIModel):
    """
    A class that represents a text generation task using the Hugging Face API.

    It inherits from the HuggingFaceModel class and overrides the call method to
    specifically work with text generation models.

    This class allows users to simply pass a text string and get a generated
    text in return.

    Args:
        api_key (str): The API key to access the Hugging Face API.

    Example:

    # >>> model = HFTextGenerationTask(api_key=token)
    # >>> text = "Once upon a time"
    # >>> generated_text = model(text, locator="gpt2")
    # >>> print(generated_text)
    'Once upon a time in a land far away...'

    """

    def __call__(self, text: str, locator="gpt2", **kwargs):
        response = self.call_api(inputs=text, model=locator, **kwargs)
        return response[0]["generated_text"]



import functools
import os
from abc import ABC, abstractmethod
from typing import Union

import diskcache
import pathway as pw
from llm_app.model_wrappers.api_clients.clients import APIClient


class _Cache:
    """A simple cache"""

    def __init__(self) -> None:
        if cache_dir := os.environ.get("PATHWAY_CACHE_DIR"):
            self.cache = diskcache.Cache(cache_dir)
        else:
            self.cache = {}

    def __call__(self, fun):
        base_name = f"{fun.__module__}_{fun.__qualname__}"

        @functools.wraps(fun)
        def wrapper(*args, **kwargs):
            key = f"{base_name}({(args, kwargs)})"
            if key not in self.cache:
                self.cache[key] = fun(*args, **kwargs)
            return self.cache[key]

        return wrapper


class BaseModel(ABC):
    def __init__(self, **kwargs):
        self.config = kwargs
        self.cache = _Cache()

    def __call__(self, text: str, **kwargs):
        raise NotImplementedError()

    def apply(
        self,
        text: Union[pw.ColumnExpression, str],
        **kwargs,
    ) -> pw.ColumnExpression:
        return pw.apply_async(self.cache(self.__call__), text=text, **kwargs)


class APIModel(BaseModel):
    def __init__(self, api_key: str, **kwargs):
        super().__init__(**kwargs)
        self.api_client = self.get_client(api_key)

    @abstractmethod
    def get_client(self, api_key: str) -> APIClient:
        pass
